use std::env;
use std::fs::File;
use std::io::{Read, Write};

use std::sync::mpsc;
use std::time::Duration;

extern crate tenquestionmarks;
use tenquestionmarks::Tenquestionmarks;

extern crate toml;

#[macro_use]
extern crate log;
extern crate env_logger;

extern crate time;

extern crate crossbeam;
extern crate notify;
use notify::{RecommendedWatcher, Watcher, RecursiveMode, DebouncedEvent};

fn init_logger () {
    let mut builder = env_logger::Builder::new();
    builder.format(|buf, record: &log::Record| {
        let t = time::now();
        writeln!(buf, "{} {}:{}: {}", time::strftime("%Y-%m-%d %H:%M:%S", &t).unwrap(), record.level(), record.module_path().unwrap_or("?"), record.args())
    }).filter(None, log::LevelFilter::Info);;

    if env::var("RUST_LOG").is_ok() {
       builder.parse(&env::var("RUST_LOG").unwrap());
    }

    builder.init();
}

fn main () {
    init_logger();

    let config_file_name = env::args().nth(1).unwrap_or("tenquestionmarks.toml".into());
    match Tenquestionmarks::from_configuration(read_config_from_file(&config_file_name)) {
        Ok(tqm) => {
            info!("tenquestionmarks initialized successfully");
            crossbeam::scope(|scope| {
                scope.spawn(|| {
                    let (notify_sender, notify_reciever) = mpsc::channel();
                    let mut watcher: RecommendedWatcher = Watcher::new(notify_sender, Duration::from_secs(2)).expect("Failed to create watcher");
                    watcher.watch(&config_file_name, RecursiveMode::NonRecursive).expect("Failed to watch config file");

                    loop {
                        if let Ok(event) = notify_reciever.recv() {
                            if let DebouncedEvent::Write(_) = event {
                                info!("Detected modified config file, reconfiguring");
                                tqm.reconfigure(&read_config_from_file(&config_file_name));
                            }
                        }
                    }
                });

                tqm.run();
            })
        },
        Err(e) => error!("Failed to initialize tenquestionmarks: {:?}", e)
    }
}

fn read_config_from_file (config_file_name: &str) -> toml::value::Table {
    if let Ok(mut file) = File::open(&config_file_name) {
        let mut contents = String::new();
        if let Err(e) = file.read_to_string(&mut contents) {
            panic!("Failed to open config file {}: {:?}", config_file_name, e);
        } else {
            match toml::from_str(&contents) {
                Ok(configuration) => {
                    info!("Loaded configuration from: {}", config_file_name);
                    configuration
                },
                Err(error) => panic!("Failed to parse config file {}: {:?}. Config file must be a valid TOML file.", config_file_name, error)
            }
        }
    } else {
        panic!("Failed to open config file! Please specify path to a config file.");
    }
}
