use irc::client::prelude::*;
use irc::client::data::command::Command;

use modules::EventLoop;
use toml::value::Table;

use event;

use std::sync::{Arc, Mutex};
use std::sync::mpsc::Receiver;
use transformable_channels::mpsc::ExtSender;

use std::fmt;
use std::fmt::{Debug, Formatter};

use crossbeam;

use thread_local::CachedThreadLocal;
use {MessageSender, Message, User, Channel};

pub struct IrcHandler {
    config: Config
}

const DEFAULT_NICK: &'static str = "tenquestionmarks";

impl IrcHandler {
    pub fn new (_: &Table, configuration: &Table) -> Box<EventLoop> {
        Box::new(IrcHandler {
            config: Config {
                nickname: Some(configuration.get("nickname")
                                            .and_then(|value| value.as_str())
                                            .unwrap_or(DEFAULT_NICK)
                                            .to_owned()),
                server: Some(configuration.get("server")
                                          .and_then(|value| value.as_str())
                                          .expect("Expected server in IRC config")
                                          .to_owned()),
                channels: Some(configuration.get("channels")
                             .and_then(|value| value.as_array())
                             .map(|value| value.to_vec())
                             .unwrap_or(vec![])
                             .into_iter()
                             .map(|value| { String::from(value.as_str().unwrap()) })
                             .collect()),
                .. Default::default()
            }
        })
    }
}

pub struct IrcMessageSender {
    irc: IrcServerWrapper,
    channel: String
}

impl MessageSender for IrcMessageSender {
    fn send_message (&self, message: &str) {
        debug!("Send message to channel {:?}: {:?}", self.channel, message);
        match self.irc.get().send_privmsg(&self.channel, message) {
            Ok(_) => { debug!("Send message succeeded"); },
            Err(err) => { error!("Send message failed: {:?}", err) }
        }
    }
}

impl Debug for IrcMessageSender {
    fn fmt (&self, formatter: &mut Formatter) -> fmt::Result {
        write!(formatter, "IrcMessageSender {{ channel: {:?} }}", self.channel)
    }
}

fn make_user (prefix: &Option<String>, server: &IrcServer) -> Option<User> {
    prefix.as_ref().and_then(|prefix| prefix.split("!").next()).map(|name| User {
        name: name.to_owned(),
        sender: Box::new(IrcMessageSender {
            irc: IrcServerWrapper::new(&server),
            channel: name.to_owned()
        })
    })
}

struct IrcServerWrapper {
    server: IrcServer,
    thread_local: CachedThreadLocal<IrcServer>
}

impl IrcServerWrapper {
    pub fn new (server: &IrcServer) -> IrcServerWrapper {
        IrcServerWrapper {
            server: server.clone(),
            thread_local: CachedThreadLocal::new()
        }
    }

    pub fn get (&self) -> &IrcServer {
        self.thread_local.get_or(|| Box::new(self.server.clone()))
    }
}

unsafe impl Sync for IrcServerWrapper {}

impl EventLoop for IrcHandler {
    fn run (&self, sender: Box<ExtSender<event::Event>>, receiver: Receiver<Arc<event::Envelope>>) {
        let server = IrcServer::from_config(self.config.clone()).unwrap();
        server.identify().unwrap();

        crossbeam::scope(|scope| {
            let server_sender = server.clone();
            scope.spawn(move || {
                loop {
                    if let Ok(envelope) = receiver.recv() {
                        if let event::Event::Message { ref message } = envelope.event {
                            if let Some(ref channel) = message.channel {
                                server_sender.send_privmsg(&channel.name, &message.content);
                            }
                        }
                    } else {
                        break;
                    }
                }
            });

            for server_message in server.iter() {
                if let Ok(irc_message) = server_message {
                    match irc_message.command {
                        Command::PRIVMSG(channel, message) => {
                            if let Some(author) = make_user(&irc_message.prefix, &server) {
                                let message = Message {
                                    author: author,
                                    content: message,
                                    channel: Some(Channel {
                                        name: channel.clone(),
                                        description: "".to_owned(),
                                        topic: "".to_owned(),
                                        sender: Box::new(IrcMessageSender {
                                            irc: IrcServerWrapper::new(&server),
                                            channel: channel.clone()
                                        })
                                    })
                                };

                                if let Err(err) = sender.send(event::Event::Message { message: message }) {
                                    error!("Error sending message event: {:?}", err)
                                }
                            }
                        },
                        _ => {}
                    }
                }
            }
        });
    }
}
